#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>

#define QLEN 8
#define MAXCHILDREN 255
#define QUIT "\\QUIT\n"
#define GREETING "AHOJ, udv a chatszerveren, mi a becses neved?\n"

void usage();
int createlistensocket(int port, int qlen);
int clientconnect(int listen_sock);
void chat_loop(int clientsock, int common, int msg);
void addchild(pid_t pid, int des);
void rmchild(pid_t pid);
int getline(int fd, char **buf, int *buflen);
void deliver(int fd);

struct child 
{ 
    pid_t pid; 
    int msgdes;
} childrendata[MAXCHILDREN];

int main(int argc, char *argv[]) 
{
    int port;
    pid_t pid;
    int listensock, clientsock;
    int children;
    int commonpipe[2], childpipe[2];
    fd_set mask, rmask;
    int nfound;
    static struct timeval timeout;
    
    if (argc!=2) usage();
    if ((port=atoi(argv[1],10))<1024) usage();
    
    for(children=0; children<MAXCHILDREN; children++) childrendata[children].pid=0;
    children=0;

    listensock=createlistensocket(port, QLEN);
    printf("Listening socket on port %d, waiting queue length %d\n", port, QLEN);
    
    if (pipe(commonpipe))
    {
	printf("pipe failed\n");
	exit(1);
    }
    
    FD_ZERO(&mask);
    FD_SET(listensock, &mask);
    FD_SET(commonpipe[0], &mask);

    while(1) 
    {
	timeout.tv_sec=0;
	timeout.tv_usec=500000;
	rmask=mask;
	nfound=select(FD_SETSIZE, &rmask, NULL, NULL, &timeout);

	//ki ert veget?
	if (children>0 && (pid=waitpid(WAIT_ANY, NULL, WNOHANG))!=0)
	{
	    children--;
	    rmchild(pid);
	}
	
	if (nfound<0) 
	{
	    if (errno==EINTR)
	    {
		printf("Interrupted system Call\n");
		continue;
	    }
	    perror("select");
	    exit(1);
	}
	
	if (nfound==0) continue;
	
	if (FD_ISSET(listensock, &rmask)  && children<MAXCHILDREN)
	{
	    if ((clientsock=clientconnect(listensock))>0)
	    {
		pipe(childpipe);
		if ((pid=fork())==0)
		{
		    /*CHILD*/
		    close(childpipe[1]);
		    close(listencok);
		    chat_loop(clientsock, dup(commonpipe[1]), childpipe[0]);
		    exit(0);
		}
		else 
		{
		    /*PARENT*/
		    children++;
		    addchild(pid, childpipe[1]);
		    close(clientsock);
		    close(childpipe[0]);
		}
	    }
	}

	//van-e uj uzenet?
	if (FD_ISSET(commonpipe[0], &rmask)) deliver(commonpipe[0]);
    }
}

/*Az osszes child-nak kikezbesiti az uzeneteket,*/
void deliver(int fd)
{
    int i, buflen=0, llen=0;
    char *buf=NULL;

    llen=getline(fd, &buf, &buflen);
    //delivering
    for(i=0;i<MAXCHILDREN;i++)
    {
	if (childrendata[i].pid!=0)
	    write(childrendata[i].msgdes, buf, llen);
    }
    
    free(buf);
}

void addchild(pid_t pid, int des)
{
    int empty=-1;
    
    while (childrendata[++empty].pid!=0);
    childrendata[empty].pid=pid;
    childrendata[empty].msgdes=des;
}

void rmchild(pid_t pid)
{
    int p=-1;
    while (childrendata[++p].pid!=pid);
    childrendata[p].pid=0;
    close(childrendata[p].msgdes);
}

int clientconnect(int listen_sock)
{
    int client_sock, addrsize;
    struct sockaddr_in client_addr;
    
    client_sock=accept(listen_sock, &client_addr, &addrsize);
    
    printf("Incomming connection opened for %s:%d\n", 
	inet_ntoa(client_addr.sin_addr.s_addr), 
	ntohs(client_addr.sin_port));

    return client_sock;
}

int createlistensocket(int port, int qlen)
{
    int listen_sock;
    struct sockaddr_in servaddr;
    
    if ((listen_sock=socket(AF_INET, SOCK_STREAM, 0))<0)
    {
	perror("socket");
	exit(1);
    }

    bzero(&servaddr, sizeof(servaddr));
    servaddr.sin_family=AF_INET;
    servaddr.sin_port=htons(port);
    servaddr.sin_addr.s_addr=htonl(INADDR_ANY);
    
    if (bind(listen_sock, (struct sockaddr *)&servaddr, sizeof(servaddr))<0)
    {
	perror("bind");
	exit(1);
    }
    
    if (listen(listen_sock, qlen)<0)
    {
	perror("listen");
	exit(1);
    }
    
    return listen_sock;
}

void chat_loop(int sock, int common, int msg)
{
    FILE *out;
    char *buf=NULL;
    char name[32];
    int buflen=0,llen;
    int quit=0,i;
    int nfound;
    fd_set mask, rmask;
    
    out=fdopen(common,"w");
    setvbuf(out, NULL, _IOLBF, 0);	//line buffered
    
    //NEM ellenorzom le, hogy van-e mar bent ilyen nevvel valaki
    write(sock, GREETING, strlen(GREETING));
    do llen=getline(sock, &buf, &buflen); while (llen<1);
    buf[llen-1]=0;
    strncpy(name,buf,31);
    name[31]='\0';
    fprintf(out, "%s logged in\n", name);
    
    FD_ZERO(&mask);
    FD_SET(msg, &mask);
    FD_SET(sock, &mask);
    
    while (!quit)
    {
	rmask=mask;
	nfound=select(FD_SETSIZE, &rmask, NULL, NULL, NULL);
	if (nfound<0)
	{
	    if (errno==EINTR) 
	    {
		printf("interrupted System Call\n");
		continue;
	    }
	    perror("select");
	    exit(1);
	}
	if (nfound==0) continue;
	
	if (FD_ISSET(msg, &rmask))
	{
	    llen=getline(msg, &buf, &buflen);
	    write(sock,buf,llen);
	}

	if (FD_ISSET(sock, &rmask))
	{
    	    getline(sock, &buf, &buflen);
	    if (strncmp(buf, QUIT, sizeof(QUIT))==0) quit=1;
	    else fprintf(out, "%s> %s", name, buf);
	}
    }
    
    fprintf(out, "%s logged out\n", name);

    fclose(out);
    close(sock);
    close(msg);
    free(buf);
}

int getline(int fd, char **buf, int *buflen)
{
    int llen=0;

    if (*buf==NULL)
    {
	*buflen=256;
	*buf=(char *)malloc(256);
    }

    do {
	llen+=read(fd, &((*buf)[llen]), 1);
	if (llen+1==(*buflen)) *buf=(char *)realloc(*buf,++(*buflen));
    } while((*buf)[llen-1]!='\n');
    
    if ((*buf)[llen-2]=='\r')
    {
      llen--;
      (*buf)[llen-1]='\n';
    }
    
    (*buf)[llen]='\0';

    return llen;
}

void usage()
{
    printf("Usage : chatserver port \n");
    exit(0);
}