#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
 
#define QLEN	8
#define GREETING "AHOJ! Irj valamit, a tobbi kliensre kikuldom.\n"

void usage();
int getline(int fd, char **buf, int *buflen);
int createlistensocket(int port, int qlen);
int clientconnect(int listen_sock);
void clientclose(int client_sock);
     
int
 main (int argc, char *argv[])
{
    int listen_sock;
    fd_set active_fd_set, read_fd_set;
    int i,j,port;
    char *buf=NULL;
    int buflen=0,msglen;
     
    if (argc!=2) usage();
    if ((port=atoi(argv[1]))<1024) usage();
    /* socket letrehozasa*/
    listen_sock = createlistensocket(port, QLEN);
    printf("Waiting connections on port %d\n",port);
    
    /* csoportok beallitasa */
    FD_ZERO (&active_fd_set);
    FD_SET (listen_sock, &active_fd_set);

    while (1)
    {
	/* varakozas az adatokra*/
        read_fd_set = active_fd_set;
        if (select (FD_SETSIZE, &read_fd_set, NULL, NULL, NULL) < 0)
        {
    	    perror ("select");
            exit (EXIT_FAILURE);
	}
 
	/* Kapcsolodasi keres */
    	if (FD_ISSET(listen_sock, &read_fd_set)
)
        {
            int client_sock;
	
	    client_sock=clientconnect(listen_sock);
            if (client_sock < 0)
 perror ("accept");
	    else 
	    {
		FD_SET(client_sock, &active_fd_set);
		write(client_sock, GREETING, strlen(GREETING));
	    }
	    FD_CLR(listen_sock, &read_fd_set);
	}

	/* Kliensek kiszolgalasa */
	for (i = 0; i < FD_SETSIZE; ++i)
	    if (FD_ISSET (i, &read_fd_set))
            {
		msglen=getline(i, &buf, &buflen);
		if (msglen==0 || strncmp(buf,"\\QUIT\n", 5)==0)
		{
            	    clientclose(i);
                    FD_CLR(i, &active_fd_set);
		}
		else
		{
		    //delivering
		    for(j=0; j< FD_SETSIZE; j++)
		    {
			if (j!=listen_sock && FD_ISSET(j, &active_fd_set))
			{
			    write(j, ">", 1);
			    write(j, buf, msglen);
			}
		    }
		}
	    }
    }
 //while(1)
}

int clientconnect(int listen_sock)
{
    int client_sock, addrsize;
    struct sockaddr_in client_addr;
    
    client_sock=accept(listen_sock, &client_addr, &addrsize);
    
    printf("Incomming connection opened for %s:%d\n", 
	inet_ntoa(client_addr.sin_addr.s_addr), 
	ntohs(client_addr.sin_port));

    return client_sock;
}

void clientclose(int client_sock)
{
    struct sockaddr_in client_addr;
    int len;
    
    len=sizeof(client_addr);
    if (getpeername(client_sock, &client_addr, &len)<0)
    {
	perror("getpeername");
	close(client_sock);
	return;
    }
    close(client_sock);
    printf("Connection closed for %s:%d\n",
	inet_ntoa(client_addr.sin_addr.s_addr), 
	ntohs(client_addr.sin_port));
}

int createlistensocket(int port, int qlen)
{
    int listen_sock;
    struct sockaddr_in servaddr;
    
    if ((listen_sock=socket(AF_INET, SOCK_STREAM, 0))<0)
    {
	perror("socket");
	exit(1);
    }

    bzero(&servaddr, sizeof(servaddr));
    servaddr.sin_family=AF_INET;
    servaddr.sin_port=htons(port);
    servaddr.sin_addr.s_addr=htonl(INADDR_ANY);
    
    if (bind(listen_sock, (struct sockaddr *)&servaddr, sizeof(servaddr))<0)
    {
	perror("bind");
	exit(1);
    }
    
    if (listen(listen_sock, qlen)<0)
    {
	perror("listen");
	exit(1);
    }
    
    return listen_sock;
}

int getline(int fd, char **buf, int *buflen)
{
    int llen=0,l;

    if (*buf==NULL)
    {
	*buflen=256;
	*buf=(char *)malloc(256);
    }

    do {
	if ((l=read(fd, &((*buf)[llen]), 1))==0) return 0;	//EOF
	llen+=l;
	if (llen+1==*buflen) *buf=(char *)realloc(*buf,++(*buflen));
    } while((*buf)[llen-1]!='\n');
    
    if ((*buf)[llen-2]=='\r')
    {
      llen--;
      (*buf)[llen-1]='\n';
    }
    
    (*buf)[llen]='\0';

    return llen;
}

void usage()
{
    printf("Usage : chatserver port \n");
    exit(0);
}
