#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>


#define QLEN 8
#define MAXTHREADS 255
#define QUIT "\\QUIT\n"
#define GREETING "AHOJ, udv a chatszerveren, mi a becses neved?\n"

int threads=0;
int clients[MAXTHREADS];

void usage();
int createlistensocket(int port, int qlen);
int clientconnect(int listen_sock);
void *chat_thread(void *param);
int getline(int fd, char **buf, int *buflen);
void deliver(char *msg);

int main(int argc, char *argv[]) 
{
    int port,i;
    int listensock, clientsock;
    pthread_t chld_thr;
    pthread_attr_t attr;
    
    for(i=0;i<MAXTHREADS;i++) clients[i]=0;

    if (argc!=2) usage();
    if ((port=atoi(argv[1],10))<1024) usage();
    
    listensock=createlistensocket(port, QLEN);
    printf("Listening socket on port %d, waiting queue length %d\n", port, QLEN);

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);    

    while(1) 
    {
	if (threads<MAXTHREADS && (clientsock=clientconnect(listensock))>0)
	{
	    /* uj szal keszitese a cliensnek */
	    pthread_create(&chld_thr, &attr, chat_thread, (void *) clientsock);
	}
    }
}

void addclient(int fd)
{
    int i=-1;
    while(clients[++i]!=0);
    clients[i]=fd;
}
void rmclient(int fd)
{
    int i=-1;
    while(clients[++i]!=fd);
    clients[i]=0;
}

void *chat_thread(void *param)
{
    char *buf=NULL, *msgbuf;
    char name[33];
    int buflen=0,llen,msgbuflen;
    int quit=0,i;
    int sock=(int) param;
    
    //NEM ellenorzom le, hogy van-e mar bent ilyen nevvel valaki
    write(sock, GREETING, strlen(GREETING));
    do llen=getline(sock, &buf, &buflen); while (llen<1);
    buf[llen-1]=0;
    strncpy(name,buf,32);
    name[32]='\0';
    msgbuf=(char *)malloc(256);
    msgbuflen=256;
    sprintf(msgbuf, "%s logged in\n", name);
    addclient(sock);
    deliver(msgbuf);
    
    while (!quit)
    {
    	llen=getline(sock, &buf, &buflen);
	if (strncmp(buf, QUIT, sizeof(QUIT))==0) quit=1;
	else 
	{
	    if (msgbuflen<llen+35) 
	    { 
		msgbuf=(char *)realloc(msgbuf, llen+35);
		msgbuflen=llen+35;
	    }
	    sprintf(msgbuf, "%s> %s", name, buf);
	    deliver(msgbuf);
	}
    }
    
    sprintf(msgbuf, "%s logged out\n", name);
    deliver(msgbuf);
    rmclient(sock);    

    close(sock);
    free(buf);
    free(msgbuf);

    pthread_exit((void *)0);
}


/*Az osszes child-nak kikezbesiti az uzeneteket,*/
void deliver(char *msg)
{
    int i;
    
    for(i=0;i<MAXTHREADS;i++)
    {
	if (clients[i]!=0)
	    write(clients[i], msg, strlen(msg));
    }
}

int clientconnect(int listen_sock)
{
    int client_sock, addrsize;
    struct sockaddr_in client_addr;
    
    client_sock=accept(listen_sock, &client_addr, &addrsize);
    
    printf("Incomming connection opened for %s:%d\n", 
	inet_ntoa(client_addr.sin_addr.s_addr), 
	ntohs(client_addr.sin_port));

    return client_sock;
}

int createlistensocket(int port, int qlen)
{
    int listen_sock;
    struct sockaddr_in servaddr;
    
    if ((listen_sock=socket(AF_INET, SOCK_STREAM, 0))<0)
    {
	perror("socket");
	exit(1);
    }

    bzero(&servaddr, sizeof(servaddr));
    servaddr.sin_family=AF_INET;
    servaddr.sin_port=htons(port);
    servaddr.sin_addr.s_addr=htonl(INADDR_ANY);
    
    if (bind(listen_sock, (struct sockaddr *)&servaddr, sizeof(servaddr))<0)
    {
	perror("bind");
	exit(1);
    }
    
    if (listen(listen_sock, qlen)<0)
    {
	perror("listen");
	exit(1);
    }
    
    return listen_sock;
}

int getline(int fd, char **buf, int *buflen)
{
    int llen=0;

    if (*buf==NULL)
    {
	*buflen=256;
	*buf=(char *)malloc(256);
    }

    do {
	llen+=read(fd, &((*buf)[llen]), 1);
	if (llen+1==(*buflen)) *buf=(char *)realloc(*buf,++(*buflen));
    } while((*buf)[llen-1]!='\n');
    
    if ((*buf)[llen-2]=='\r')
    {
      llen--;
      (*buf)[llen-1]='\n';
    }
    
    (*buf)[llen]='\0';

    return llen;
}

void usage()
{
    printf("Usage : chatserver port \n");
    exit(0);
}